classdef ooGmsh4 < handle  
% CLASSDEF ooGmsh4:
%   
%
% Copyright (C) 2017-2019 F. Cuvelier
  properties
    q=[];
    nq=[];
    dim=[];
    toGlobal=[];
    meshfile=[];
    partitionnedfile=false;
    MeshFormat=[];  
    PhysicalNames=[];
    Entities=[];
    PartitionedEntities=[];
    Nodes=[];
    Elements=[];
    PeriodicLinks=[];
  end
  
  methods
  
    function self=ooGmsh4(gmsh_file,varargin)
    % ooGmsh4 constructor, create an ooGmsh4 object from a gmsh mesh file.
    %   
    % USAGE:
    %   Gh=ooGmsh4(gmsh_file);
    %
    % INPUT:
    %   gmsh_file: name (string) of the gmsh mesh file to read.
    %
    % OUTPUT:
    %   Gh: an ooGmsh4 object.
    %
    % Copyright (C) 2017-2019 F. Cuvelier
      [fid,message]=fopen(gmsh_file,'r');
      if ( fid == -1 )
          error([message,' : ',gmsh_file]);
      end
      self.meshfile=gmsh_file;
      self.MeshFormat=fc_oogmsh.msh.read_MeshFormat(fid);
      assert( ismember(self.MeshFormat.version,{'4','4.1'}) , 'Needs version 4 or 4.1 of msh file format! Found version %s',self.MeshFormat.version )
      switch self.MeshFormat.version
        case '4.1'
          self.PhysicalNames=fc_oogmsh.msh4_1.read_PhysicalNames(fid,varargin{:});
          self.Entities=fc_oogmsh.msh4_1.read_Entities(fid,varargin{:});
          self.PartitionedEntities=fc_oogmsh.msh4_1.read_PartitionedEntities(fid,varargin{:});
          self.Nodes=fc_oogmsh.msh4_1.read_Nodes(fid,varargin{:});
          self.Elements=fc_oogmsh.msh4_1.read_Elements(fid,varargin{:});
          self.PeriodicLinks=fc_oogmsh.msh4_1.read_Periodic(fid,varargin{:});
        case '4'
          self.PhysicalNames=fc_oogmsh.msh4_0.read_PhysicalNames(fid,varargin{:});
          self.Entities=fc_oogmsh.msh4_0.read_Entities(fid,varargin{:});
          self.PartitionedEntities=fc_oogmsh.msh4_0.read_PartitionedEntities(fid,varargin{:});
          self.Nodes=fc_oogmsh.msh4_0.read_Nodes(fid,varargin{:});
          self.Elements=fc_oogmsh.msh4_0.read_Elements(fid,varargin{:});
          self.PeriodicLinks=fc_oogmsh.msh4_0.read_Periodic(fid,varargin{:});
      end
      fclose(fid);
      self.partitionnedfile=~isempty(self.PartitionedEntities);
      [self.q,self.toGlobal]=get_q(self);
      if sum(abs(self.q(3,:)))==0,self.q=self.q(1:2,:);end
      [self.dim,self.nq]=size(self.q);
    end % ooGmsh4 constructor
    
    function [q,tags]=get_q(self,varargin)
%      dimEntities=[self.Nodes.EntityBlocks(:).entityDim];
%      p = inputParser;
%      p.addParamValue('dimEntity',max(dimEntities),@(x) ismember(x,dimEntities));
%      p.parse(varargin{:});
%      dimEntity=p.Results.dimEntity;
%      idx=find(dimEntities==dimEntity);
      tags=horzcat(self.Nodes.EntityBlocks(:).nodeTags);
      q=horzcat(self.Nodes.EntityBlocks(:).Nodes);
    end
    
    function me=get_me(self,i)
      me=self.Elements.EntityBlocks(i).nodeTags;
    end
    
    function disp(self)
      fc_tools.comp.dispobject(self)
    end
    
    function info(self,varargin)
      fc_tools.comp.infoobject(self,varargin{:})
    end
    
    function tags=get_ElementaryTags(Gh,type)
      idx=find(Gh.Elements.ElementTypes==type);
      if isempty(idx), tags=[];return;end
      tags=[Gh.Elements.EntityBlocks(idx).entityTag];
    end
    
    function tags=get_PhysicalTags_entityDim(Gh,entityDim)
      switch entityDim
        case 0 % Points
          tags=unique([Gh.Entities.Points(:).PhysicalTags]);
        case 1 % Curves
          tags=unique([Gh.Entities.Curves(:).PhysicalTags]);
        case 2 % Surfaces
          tags=unique([Gh.Entities.Surfaces(:).PhysicalTags]);
        case 3 % Volumes
          tags=unique([Gh.Entities.Volumes(:).PhysicalTags]);
      end
    end
    
    function sE=get_subEntities(Gh,EltType)
      sE=[];
      elt=fc_oogmsh.gmsh.elm_type_desc(EltType);
      assert( ~isempty(elt) )
      idx=find(Gh.Elements.ElementTypes==EltType);
      if isempty(idx),return;end
      EltsTags=[Gh.Elements.EntityBlocks(idx).entityTag];

      entityDim=elt.d;
      switch entityDim
        case 0 % Points
          sE=Gh.Entities.Points;
        case 1 % Curves
          sE=Gh.Entities.Curves;
        case 2 % Surfaces
          sE=Gh.Entities.Surfaces;
        case 3 % Volumes
          sE=Gh.Entities.Volumes;
      end

      sEtags=[sE(:).tag]; 
      I=ismember(sEtags,EltsTags);
      sE=sE(I);
    end
    
    function tags=get_PhysicalTags(Gh,EltType)
      sE=Gh.get_subEntities(EltType);
      if isempty(sE), tags=[];else, tags=unique([sE.PhysicalTags]);end
    end
    
    function varargout=get_localmesh_ElementaryTag(self,EltType,EltTag,varargin)
      assert( ismember(nargout,[2,3]) )
      % returns  [q,me] or [q,me,toGlobal]
      p = inputParser;
      p.addParamValue('dim',3,@(x) ismember(x,[2,3]));
      p.parse(varargin{:});
      dim=p.Results.dim;
      me=self.get_me_ElementaryTag(EltType,EltTag);
      %[q,qtags]=self.get_q();
      [q,me,toGlobal]=fc_oogmsh.msh.FromGlobalToLocal(self.q,me,self.toGlobal);
      varargout{1}=q;
      varargout{2}=me;
      if nargout==3,varargout{3}=toGlobal;end  
    end     
    
    function varargout=get_localmesh_PhysicalTag(self,EltType,PhysTag,varargin)
      assert( ismember(nargout,[2,3]) )
      % returns  [q,me] or [q,me,toGlobal]
      p = inputParser;
      p.addParamValue('dim',3,@(x) ismember(x,[2,3]));
      p.parse(varargin{:});
      dim=p.Results.dim;
      me=self.get_me_PhysicalTag(EltType,PhysTag);
      [q,me,toGlobal]=fc_oogmsh.msh.FromGlobalToLocal(self.q,me,self.toGlobal);
      varargout{1}=q;
      varargout{2}=me;
      if nargout==3,varargout{3}=toGlobal;end  
    end
    
%      function varargout=get_localmesh_PhysicalTag_dimEntity(self,dimEntity,PhysTag,varargin)
%        assert( ismember(nargout,[2,3]) )
%        % returns  [q,me] or [q,me,toGlobal]
%        p = inputParser;
%        p.addParamValue('dim',3,@(x) ismember(x,[2,3]));
%        p.parse(varargin{:});
%        dim=p.Results.dim;
%        me=self.get_me_PhysicalTag_dimEntity(dimEntity,PhysTag);
%        %[q,qtags]=self.get_q();
%        [q,me,toGlobal]=fc_oogmsh.msh.FromGlobalToLocal(self.q,me,self.toGlobal);
%  %        if dim==2
%  %          assert( sum(abs(q(3,:)))==0, 'Not a 2D mesh, cannot force dimension to 2')
%  %          q=q(1:2,:);
%  %        end
%        varargout{1}=q;
%        varargout{2}=me;
%        if nargout==3,varargout{3}=toGlobal;end  
%      end

    function me=get_me_ElementaryTag(Gh,type,tag) % tag is an ElementaryTag
      idx=find(Gh.Elements.ElementTypes==type);
      if isempty(idx), me=[];return;end
      i=find([Gh.Elements.EntityBlocks(idx).entityTag]==tag);
      if isempty(i), me=[];return;end
      assert(isscalar(i))
      me=Gh.Elements.EntityBlocks(idx(i)).nodeTags;
    end
    
%      function me=get_me_PhysicalTag_dimEntity(self,dimEntity,phtag)
%      % only works with order 1 simplices.
%        switch dimEntity 
%          case 1 % Curves
%            Entities=self.Entities.Curves;
%            elementtype=1; % line (see fc_oogmsh.gmsh.elm_type_list function)
%          case 2 % Surfaces
%            Entities=self.Entities.Surfaces;
%            elementtype=2; % triangle (see fc_oogmsh.gmsh.elm_type_list function)
%          case 3 % Volumes
%            Entities=self.Entities.Volumes;
%            elementtype=4; %(see fc_oogmsh.gmsh.elm_type_list function)
%        end
%        numPTs=[Entities(:).numPhysicalTags];
%        idx=find(numPTs>0);
%        Entities=Entities(idx);
%        %phtags=horzcat(Entities(:).PhysicalTags);
%        %idx=find(phtags==phtag);
%        phtags={Entities(:).PhysicalTags}; 
%        idx=cellfun(@(x) ismember(phtag,x) , phtags);
%        tags=horzcat(Entities(idx).tag); 
%        idx=find(self.Elements.ElementTypes==elementtype); % 1: Lines
%        Etags=[self.Elements.EntityBlocks(idx).entityTag];
%        [~,~,ib]=intersect(tags,Etags); % Etags(ib)==tags
%        Idx=idx(ib);
%        me=[self.Elements.EntityBlocks(Idx).nodeTags];
%      end
    
    function me=get_me_PhysicalTag(Gh,EltType,phtag)
      elt=fc_oogmsh.gmsh.elm_type_desc(EltType);
      idx=find(Gh.Elements.ElementTypes==EltType);
      sE=Gh.get_subEntities(EltType);
      J=cellfun(@(x) ismember(phtag,x),  {sE(:).PhysicalTags});
      tags=[sE(J).tag];
      i=ismember([Gh.Elements.EntityBlocks(idx).entityTag],tags);
      me=[Gh.Elements.EntityBlocks(idx(i)).nodeTags];
    end

    
  end % methods
end % classdef



